package sets;


import java.util.concurrent.atomic.AtomicMarkableReference;


// lock-free node, which supports operations
// to update next and valid together, atomically
class LockFreeNode<T> extends BasicNode<T>
{
    // reference (next node in chain) and mark (validity of _current_ node)
    private AtomicMarkableReference<Node<T>> nextValid;

    @Override
    public Node<T> next() {
        return nextValid.getReference();
    }

    @Override
    public void setNext(Node<T> next) {
        throw new UnsupportedOperationException("Method setNext() is not implemented.");
    }

    @Override
    public boolean valid() {
        return nextValid.isMarked();
    }

    @Override
    public boolean setValid() {
        return nextValid.attemptMark(next(), true);
    }

    @Override
    public boolean setInvalid() {
		  Node<T> next = next();
        return nextValid.compareAndSet(next, next, true, false);
    }

    @Override
    public boolean setNextIfValid(Node<T> expectNext, Node<T> newNext) {
        return nextValid.compareAndSet(expectNext, newNext, true, true);
    }

    // next and valid fields of node, retrieved atomically
    public Node<T> nextValid() {
        boolean[] valid = { true };  // to get value of valid
        Node<T> next = nextValid.get(valid);  // get next (returned) and valid (written to array)
        // of current node
        return new FrozenNode<>(next, valid[0]);
    }

    LockFreeNode(T item) {
        super(item);
    }

    LockFreeNode(int key) {
        super(key);
        nextValid = new AtomicMarkableReference<Node<T>>(null, true);
    }

    // node with given item, pointing to succ
    LockFreeNode(T item, Node<T> succ) {
        super(item);
        nextValid = new AtomicMarkableReference<Node<T>>(succ, true);
    }

    // node with given key, pointing to succ
    LockFreeNode(int key, Node<T> succ) {
        super(key);
        nextValid = new AtomicMarkableReference<Node<T>>(succ, true);
    }
}
